import uuid
import pandas

from .mlmodel import MLModel
from .plot import Plot
from .ilogger import ILogger
from .api_interface import ApiInterface
from .pipeline import PipelineStep

from .mlmodel import MLModelConfig
from .script import ScriptConfig
from .evaluation import Evaluation

from .template import Template
from .twinregistration import TwinRegistration

from datetime import datetime


class Context(ILogger):
    """
    Context defines all properties, dataframes, models and other data passed through pipeline steps.
    - properties are accessible on properties property.
    - output dataframe(s) use return statement.
    - input dataframe(s) get them by parameters or use dataframe property directly.
    - for model use the set_model statement : support only one model per script.
    - for plot use the set_model statement : support only one plot per script.
    - dataframes, models and additional plots used within the pipelines can also be accessed, read-only.
    :ivar wizata_dsapi.ApiInterface api: api instance with limited functionality accessible during runner execution.
    :ivar dict config: all Wizata custom.ini configuration keys.
    :ivar pandas.Dataframe dataframe: single instance of pandas Dataframe inputted as parameter.
    :ivar dict dataframes: dict of all dataframes parameters and generated by previously executed steps.
    :ivar dict datapoints: wizata_dsapi.Datapoint for each name (key) used in queries.
    :ivar uuid.UUID execution_id: Execution ID currently processed.
    :ivar grafana_client.GrafanaApi grafana_api: instance of grafana api.
    :ivar dict models: dict of all models generated by previously executed steps.
    :ivar datetime now: Now timestamp as used by the pipeline.
    :ivar uuid.UUID pipeline_id: Pipeline ID associated with the ongoing execution.
    :ivar dict plots: dict of all plots generated by previously executed steps.
    :ivar dict properties: properties dictionary containing useful information such as variables.
    :ivar wizata_dsapi.PipelineStep step: Currently processed step definition.
    :ivar wizata_dsapi.Template template: Template if set on the pipeline.
    :ivar wizata_dsapi.TwinRegistration registration: Registration associated with the pipeline and the twin currently processed.
    :ivar list warnings: all warnings and error messages.
    """

    def __init__(self,
                 pipeline_id: uuid.UUID = None,
                 execution_id: int = None,
                 template: Template = None,
                 registration: TwinRegistration = None,
                 now: datetime = None,
                 step: PipelineStep = None,
                 dataframe=None,
                 properties=None):
        self.pipeline_id = pipeline_id
        self.execution_id = execution_id
        self.template = template
        self.registration = registration
        self.now = now
        self.step = step

        self.models = {}
        self.dataframes = {}
        self.datapoints = {}
        if properties is None:
            properties = {}
        self.properties = properties
        self.plots = {}

        self._logger = None

        self.__key_mapping = None
        self.__primary_key = None
        self.__temporary_dataframe = None

        self.__new_plot = None
        self.__new_model = None

        self.grafana_api = None
        self.config = None

        self.warnings = []

        self._api = None
        self.dataframe = dataframe

    def write_log(self, message: str, level: int = 7):
        """
        write log in console (only in experiment mode) - if running locally it will print in the console.
        :param message: message to write.
        :param level: from 7=DEBUG, 6=INFO, 3=ERROR to 1=CRITICAL
        """
        if self._logger is None:
            print(message)
        else:
            self._logger.write_log(message=message, level=level)

    @property
    def api(self) -> ApiInterface:
        """
        Accessible DS API functionalities within runner context.
        """
        return self._api

    @api.setter
    def api(self, api: ApiInterface):
        self._api = api

    @property
    def dataframe(self):
        """
        dataframe - use for single input only - for output use return statement
        """
        if self.__primary_key is not None and self.__key_mapping is not None and self.__primary_key in self.__key_mapping and self.__key_mapping[self.__primary_key] in self.dataframes:
            return self.dataframes[self.__key_mapping[self.__primary_key]]
        elif self.__temporary_dataframe is not None:
            return self.__temporary_dataframe
        else:
            return None

    @dataframe.setter
    def dataframe(self, df: pandas.DataFrame):
        if self.__primary_key is not None and self.__key_mapping is not None and self.__primary_key in self.__key_mapping and self.__key_mapping[self.__primary_key] in self.dataframes:
            self.dataframes[self.__key_mapping[self.__primary_key]] = df
        else:
            self.__temporary_dataframe = df

    def append(self, key: str, obj, overwrite: bool = True):
        """
        append an object (pandas.Dataframe or any properties)
        :param key: dictionary identifier - name inside your pipeline.
        :param obj: ML Model, Dataframe or any properties (must be JSON serializable type).
        :param overwrite: by default - allow modifying an existing object. can be set to false.
        """
        if key is None or key == "":
            raise KeyError('please provide a non empty key.')

        if isinstance(obj, pandas.DataFrame):
            if not overwrite and key in self.dataframes:
                raise KeyError(f'cannot overwrite existing dataframe in context with key {key}')
            self.dataframes[key] = obj
        else:
            if not overwrite and key in self.properties:
                raise KeyError(f'cannot overwrite existing properties in context with key {key}')
            self.properties[key] = obj

    def get(self, key: str):
        """
        get key from either dataframes, models, plots or properties.
        :return: None if not found.
        """
        if key in self.dataframes:
            return self.dataframes[key]
        elif key in self.models:
            return self.models[key]
        elif key in self.plots:
            return self.plots[key]
        elif key in self.properties:
            return self.properties[key]
        else:
            return None

    def set_plot(self, figure, name="Unkwown"):
        """
        set plot to be added to the context.
        :param figure: Plotly figure.
        :param name: Name of the plot.
        :return: Plot object prepared.
        """
        plot = Plot()
        plot.name = name
        plot.figure = figure
        self.__new_plot = plot
        return plot

    def get_plot(self) -> Plot:
        """
        get plot set to be added to the context.
        """
        return self.__new_plot

    def show(self):
        """
        return plot figure to be shown in a notebook.
        :return: plot figure
        """
        if self.__new_plot is not None:
            return self.__new_plot.figure

    def set_model(self, trained_model, input_columns, output_columns=None, has_anomalies=False, scaler=None):
        """
        set model to be added to the context.
        :param trained_model: Trained Model to be stored as a pickled object.
        :param list input_columns: List of str defining input columns to call the model (df.columns)
        :param list output_columns: List of output columns - Optional as can be detected automatically during validation.
        :param bool has_anomalies: False by default, define if the model set anomalies
        :param scaler: Scaler to be stored if necessary.
        :return: ML Model object prepared.
        """
        ml_model = MLModel()
        ml_model.trained_model = trained_model
        ml_model.scaler = scaler
        ml_model.input_columns = input_columns
        ml_model.output_columns = output_columns
        ml_model.has_anomalies = has_anomalies
        self.__new_model = ml_model
        return ml_model

    def get_model(self) -> MLModel:
        """
        get model to be added to the context.
        """
        return self.__new_model

    def reset(self):
        """
        reset context between step execution - remove all step info, but keep all data.
        """
        self.__new_plot = None
        self.__new_model = None
        self.__key_mapping = None
        self.__primary_key = None
        self.__temporary_dataframe = None
        self.step = None

    def _set_mapping(self, key_mapping: dict, primary_key: str):
        """
        set step mapping between pipeline names and step name
        """
        self.__key_mapping = key_mapping
        self.__primary_key = primary_key

    def current_dataframes(self) -> dict:
        """
        - current dataframes a dictionary with all current named dataframes specific for this script.
        - dataframes contains all accessible dataframes for the pipeline mapped.
        - single dataframe context is accessible with context.dataframe and is not named
        """
        if self.__key_mapping is not None:
            dataframes = {}
            for key in self.__key_mapping:
                dataframes[self.__key_mapping[key]] = self.dataframes[key]
            return dataframes
        else:
            return self.dataframes

    def get_script_config(self) -> ScriptConfig:
        """
        extract script configuration from the context.
        """
        if self.step is None:
            raise ValueError(f'there is no step defined in the context.')
        if self.step.config is None:
            raise ValueError(f'there is no step config defined in the context.')
        if not isinstance(self.step.config, ScriptConfig):
            raise TypeError(f'step is not configured as a script/plot step')
        return self.step.config

    def get_model_config(self) -> MLModelConfig:
        """
        extract model configuration from the context.
        """
        if self.step is None:
            raise ValueError(f'there is no step defined in the context.')
        if self.step.config is None:
            raise ValueError(f'there is no step config defined in the context.')
        if not isinstance(self.step.config, MLModelConfig):
            raise TypeError(f'step is not configured as a model step')
        return self.step.config

    def evaluate(self, model_key: str, properties: dict, metrics: dict):
        """
        :param model_key: model key as designed on the model config target by this evaluation.
        :param properties: properties key/value relevant to this model training.
        :param metrics: metrics results of the evaluation.
        :return: evaluation entity created.
        """
        if self.execution_id is None:
            raise ValueError(f'this context is not running within an execution (execution_id is None).')
        if model_key not in self.models:
            raise RuntimeError(f'the model {model_key} is not present within the context.')
        model: MLModel = self.models[model_key]
        evaluation = Evaluation(
            execution_id=self.execution_id,
            model_id=model.model_id,
            properties=properties,
            metrics=metrics
        )
        if self.api is None:
            raise RuntimeError(f'no api is set on this context.')
        self.api.create(evaluation)
        return evaluation



import uuid
import json
from .api_dto import ApiDto


class Trigger(ApiDto):
    """
    trigger defines how pipelines are automatically executed.

    :ivar uuid.UUID trigger_id: technical id of the trigger.
    :ivar int interval: interval in ms between two execution.
    :ivar int delay: set a delay from 00:00:00.000ms interval alignment as ms.
    :ivar uuid.UUID pipeline_id: set a pipeline id to trigger.
    :ivar uuid.UUID template_id: set a template id to trigger.
    :ivar dict properties: set a dict of properties to pass when triggering the pipeline.
    :ivar list twin_ids: set a list of twin ids on which executed the pipelines (must be registered on the pipeline).
    :ivar uuid.UUID createdById: unique identifier of creating user.
    :ivar int createdDate: timestamp of created date.
    :ivar uuid.UUID updatedById: unique identifier of updating user.
    :ivar int updatedDate: timestamp of updated date.
    """

    @classmethod
    def route(cls):
        return "triggers"

    @classmethod
    def from_dict(cls, data):
        obj = Trigger()
        obj.from_json(data)
        return obj

    def __init__(self,
                 trigger_id: uuid.UUID = None,
                 interval: int = None,
                 delay: int = None,
                 pipeline_id: uuid.UUID = None,
                 pipeline_image_id: str = None,
                 template_id: uuid.UUID = None,
                 properties: dict = None,
                 twin_ids: list = None):
        if trigger_id is None:
            trigger_id = uuid.uuid4()
        self.trigger_id = trigger_id
        self.interval = interval
        self.delay = delay
        self.pipeline_id = pipeline_id
        self.pipeline_image_id = pipeline_image_id
        self.template_id = template_id
        self.properties = properties
        self.twin_ids = twin_ids
        self.createdById = None
        self.createdDate = None
        self.updatedById = None
        self.updatedDate = None

    def api_id(self) -> str:
        return str(self.trigger_id).upper()

    def endpoint(self) -> str:
        return "ExecutionTriggers"

    def from_json(self, obj):
        if "id" in obj.keys():
            self.trigger_id = uuid.UUID(obj["id"])

        if "interval" in obj.keys() and obj["interval"] is not None:
            self.interval = int(obj["interval"])
        if "delay" in obj.keys() and obj["delay"] is not None:
            self.delay = int(obj["delay"])

        if "pipelineId" in obj.keys() and obj["pipelineId"] is not None:
            self.pipeline_id = uuid.UUID(obj["pipelineId"])
        if "pipelineImageId" in obj.keys() and obj["pipelineImageId"] is not None:
            self.pipeline_image_id = str(obj["pipelineImageId"])
        if "templateId" in obj.keys() and obj["templateId"] is not None:
            self.template_id = uuid.UUID(obj["templateId"])

        if "jsonProperties" in obj.keys():
            if isinstance(obj["jsonProperties"], str) and obj["jsonProperties"] != "":
                self.properties = json.loads(obj["jsonProperties"])
            else:
                self.properties = obj["jsonProperties"]

        if "twinIds" in obj.keys() and isinstance(obj["twinIds"], list):
            self.twin_ids = []
            for twinId in obj["twinIds"]:
                self.twin_ids.append(uuid.UUID(twinId))

        if "createdById" in obj.keys() and obj["createdById"] is not None:
            self.createdById = obj["createdById"]
        if "createdDate" in obj.keys() and obj["createdDate"] is not None:
            self.createdDate = obj["createdDate"]
        if "updatedById" in obj.keys() and obj["updatedById"] is not None:
            self.updatedById = obj["updatedById"]
        if "updatedDate" in obj.keys() and obj["updatedDate"] is not None:
            self.updatedDate = obj["updatedDate"]

    def to_json(self, target: str = None):
        obj = {
            "id": str(self.trigger_id)
        }

        if self.interval is not None:
            obj["interval"] = self.interval
        if self.delay is not None:
            obj["delay"] = self.delay

        if self.template_id is not None:
            obj["templateId"] = str(self.template_id)

        if self.pipeline_id is not None:
            obj["pipelineId"] = str(self.pipeline_id)
        if self.pipeline_image_id is not None:
            obj["pipelineImageId"] = str(self.pipeline_image_id)

        if self.properties is not None:
            obj["jsonProperties"] = json.dumps(self.properties)

        if self.twin_ids is not None and len(self.twin_ids) > 0:
            obj["twinIds"] = []
            for twin_id in self.twin_ids:
                obj["twinIds"].append(str(twin_id))

        if self.createdById is not None:
            obj["createdById"] = str(self.createdById)
        if self.createdDate is not None:
            obj["createdDate"] = str(self.createdDate)
        if self.updatedById is not None:
            obj["updatedById"] = str(self.updatedById)
        if self.updatedDate is not None:
            obj["updatedDate"] = str(self.updatedDate)

        return obj



import wizata_dsapi

import pandas
import numpy

import sklearn
import sklearn.linear_model
import sklearn.ensemble


def extract_target_feat(context: wizata_dsapi.Context, single: bool = True):
    """
    return a list of target_feat columns names if not single value or the single value target feat name
    raise an error if configuration mismatch
    """

    if "target_feat" not in context.properties:
        raise ValueError(f"training script requires a proper target_feat")

    target_feat = context.properties["target_feat"]
    if isinstance(target_feat, str):
        if single:
            return target_feat
        else:
            return [target_feat]
    elif isinstance(target_feat, list):
        if single:
            if len(target_feat) == 1:
                return target_feat[0]
            else:
                raise ValueError(f"expecting only one target_feat but found {len(target_feat)}")
        else:
            return [target_feat]
    else:
        raise TypeError(f'target_feat must be a str or a list of str but found {target_feat.__class__.__name__}')


def linear_regression(context: wizata_dsapi.Context):
    """
    generic linear regression
    - expects a valid single value target_feat
    """
    df = context.dataframe

    model_config = context.get_model_config()
    if not model_config.has_target_feat():
        raise ValueError(f'linear_regression requires a target feat')
    target_feat_name = context.properties["target_feat"]

    x = df.drop(columns=[target_feat_name])
    y = df[target_feat_name]

    model = sklearn.linear_model.LinearRegression()
    model.fit(x, y)

    context.set_model(model, x.columns)


def logistic_regression(context: wizata_dsapi.Context):
    """
    generic linear regression
    - expects a valid single value target_feat
    """
    df = context.dataframe

    model_config = context.get_model_config()
    if not model_config.has_target_feat():
        raise ValueError(f'logistic_regression requires a target feat')
    target_feat_name = context.properties["target_feat"]

    x = df.drop(columns=[target_feat_name])
    y = df[target_feat_name]

    model = sklearn.linear_model.LogisticRegression()
    model.fit(x, y.astype(int))

    context.set_model(model, x.columns)


def isolation_forest(context: wizata_dsapi.Context):
    """
    isolation forest anomaly detection
    - expects a no target feat model step
    """

    model_config = context.get_model_config()
    if model_config.has_target_feat():
        raise ValueError(f'isolation_forest does not requires a target feat')

    try:
        if context.properties['sensitivity'] is None:
            raise KeyError("sensitivity is none")
        sensitivity = int(context.properties['sensitivity'])
        sensitivities = [0.05, 0.15, 0.25, 0.35, 0.4]
        contamination = sensitivities[sensitivity - 1]
    except Exception as e:
        raise ValueError(f'cannot extract sensitivity integer from 0 to 4 due to {e}')

    df = context.dataframe.copy()
    model = sklearn.ensemble.IsolationForest(contamination=contamination)
    df['isolation_forest_predict'] = model.fit_predict(df)
    context.set_model(model, df.columns)
    return df


def gradiant_boost_classifier(context: wizata_dsapi.Context):
    """
    gradiant_boost_classifier
    :param context:
    :return:
    """
    df = context.dataframe

    model_config = context.get_model_config()
    if not model_config.has_target_feat():
        raise ValueError(f'gradiant_boost_classifier requires a target feat')
    target_feat_name = context.properties["target_feat"]

    x = df.drop(columns=[target_feat_name])
    y = df[target_feat_name]

    model = sklearn.ensemble.GradientBoostingClassifier(random_state=0).fit(x, y)
    context.set_model(model, df.columns)









import sys


def get_streamlit_token():
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "auth_token" in st.query_params:
            auth_token = st.query_params["auth_token"]
            return auth_token


def get_streamlit_domain():
    if 'streamlit' in sys.modules:
        st = sys.modules['streamlit']
        if st.query_params is not None and "dsapi" in st.query_params:
            domain = st.query_params["dsapi"]
            return domain



from setuptools import setup
from distutils.util import convert_path

main_ns = {}
ver_path = convert_path('wizata_dsapi/version.py')
with open(ver_path) as ver_file:
    exec(ver_file.read(), main_ns)

core_dependencies = [
    'pandas==1.5.3',
    'numpy==1.26.4',
    'dill==0.3.6',
    'protobuf==3.20.3',
    'msal>=1.24.0',
    'joblib==1.2.0',
    'requests==2.28.2',
    'setuptools==67.6.0',
    'plotly==5.13.1'
]

ml_dependencies = [
    'matplotlib==3.7.1',
    "tensorflow==2.15.0; sys_platform != 'darwin' or platform_machine != 'arm64'",
    "tensorflow-macos==2.15.0; sys_platform == 'darwin' and platform_machine == 'arm64'",
    "keras==2.15.0; sys_platform != 'darwin' or platform_machine != 'arm64'",
    "keras==2.15.0; sys_platform == 'darwin' and platform_machine == 'arm64'",
    'tensorflow_probability==0.15.0',
    'scikit-learn==1.2.2',
    'adtk==0.6.2',
    'scipy==1.10.1',
    'xgboost==1.7.4',
    'u8darts==0.25.0',
    'optuna==3.3.0',
    'explainerdashboard==0.4.2.1',
    'ipywidgets==8.0.4',
    'kaleido==0.2.1',
    'pytest==7.2.2',
    'pytest-cov==4.0.0',
    'shapely==2.0.1',
    'pyodbc==4.0.35'
]
all_dependencies = core_dependencies + ml_dependencies

setup(
    name='wizata_dsapi',
    version=main_ns['__version__'],
    description='Wizata Data Science Toolkit',
    author='Wizata S.A.',
    author_email='info@wizata.com',
    packages=['wizata_dsapi',
              'wizata_dsapi.plots',
              'wizata_dsapi.scripts',
              'wizata_dsapi.models'],
    install_requires=core_dependencies,
    extras_require={
        'all': all_dependencies
    },
)



import json
import uuid
from .api_dto import ApiDto
from enum import Enum


def get_bool(obj, name: str):
    if isinstance(obj[name], str) and obj[name].lower() == "false":
        return False
    else:
        return bool(obj[name])


class MLModelConfig(ApiDto):
    """
    a model config defines execution properties within a pipeline.
    usually to define how a pipeline should train and predict with your model.

    :ivar bool by_twin: define if pipeline need to train and use different model by twin item.
    :ivar bool by_property: define if pipeline need to train and use different model based on 'property_name'.
    :ivar list features: datapoint list to refine columns if necessary.
    :ivar str function: name of the function used instead of 'predict' on model inference.
    :ivar str model_key: key of the model to store (or use story property if dynamic).
    :ivar str model_type: reserved for 'mlflow' defines model flavor such as 'sklearn' or 'pyfunc'.
    :ivar str model_alias: reserved for 'mlflow' set the alias to target and use inside a pipeline.
    :ivar dict properties_mapping: dict to map properties expected by the script with the properties from the context.
    :ivar str property_name: define which property is looked for when using 'by_property'.
    :ivar str source: define name of model storage to use 'wizata'.
    :ivar str|list target_feat: target feature name(s) to remove from df in 'predict' mode.
    :ivar str train_script: name of function referencing the script to train the model.
    :ivar float train_test_split_pct: percentage repartition to split the data for training and scoring.
    :ivar str train_test_split_type: type of splitting desired to split the train dataframe.
    :ivar bool output_append: true - default - append output to input dataframe. false to retrieve only 'predict' column(s).
    :ivar list output_columns_names: name list to rename columns inside output dataframe.
    :ivar str output_prefix: set a prefix to put before output column names (default or custom).
    """

    def __init__(self,
                 train_script=None,
                 train_test_split_pct: float = 1.0,
                 train_test_split_type: str = "ignore",
                 function: str = "predict",
                 properties_mapping=None,
                 target_feat=None,
                 output_append: bool = True,
                 output_columns_names: list = None,
                 output_prefix: str = None,
                 features: list = None,
                 model_key: str = None,
                 model_type: str = None,
                 model_alias: str = None,
                 by_twin: bool = False,
                 by_property: bool = False,
                 property_name: str = None,
                 source: str = "wizata"
                 ):

        # training
        self.train_script = train_script
        self.train_test_split_pct = train_test_split_pct
        self.train_test_split_type = train_test_split_type
        self.function = function
        self.properties_mapping = properties_mapping

        # features management
        self.target_feat = target_feat
        self.features = features
        self.output_columns_names = output_columns_names
        self.output_append = output_append
        self.output_prefix = output_prefix

        # key and identification
        self.model_key = model_key
        self.by_twin = by_twin
        self.by_property = by_property
        self.property_name = property_name

        # source
        self.source = source
        self.model_type = model_type
        self.model_alias = model_alias

    def from_json(self, obj):

        # Managed deprecated fields
        if "model_key_type" in obj.keys() and obj["model_key_type"] is not None:
            if obj["model_key_type"] == "template":
                self.by_twin = True
            elif obj["model_key_type"] == "variable":
                self.by_property = True
                if "model_key" not in obj.keys() or obj["model_key"] is None:
                    raise KeyError('model_key must be declared in the config.')
                self.property_name = obj["model_key"]

        # training info
        if "train_script" in obj.keys() and obj["train_script"] is not None:
            self.train_script = obj["train_script"]
        if "train_test_split_pct" in obj.keys() and obj["train_test_split_pct"] is not None:
            self.train_test_split_pct = float(obj["train_test_split_pct"])
        if "train_test_split_type" in obj.keys() and obj["train_test_split_type"] is not None:
            self.train_test_split_type = obj["train_test_split_type"]
        if "function" in obj.keys() and obj["function"] is not None:
            self.function = obj["function"]
        if "properties_mapping" in obj:
            self.properties_mapping = obj["properties_mapping"]

        # features info
        if "target_feat" in obj.keys() and obj["target_feat"] is not None:
            if isinstance(obj["target_feat"], str) or isinstance(obj["target_feat"], list):
                self.target_feat = obj["target_feat"]
            else:
                raise ValueError(f'target_feat should be a str or a list with columns name to remove.')

        if "features" in obj.keys() and obj["features"] is not None:
            self.features = obj["features"]

        # output management
        if "output_append" in obj.keys():
            self.output_append = get_bool(obj, name="output_append")

        if "output_prefix" in obj.keys() and obj["output_prefix"] is not None:
            self.output_prefix = obj["output_prefix"]
        if "output_columns_names" in obj.keys() and obj["output_columns_names"] is not None:
            if isinstance(obj["output_columns_names"], str):
                self.output_columns_names = [obj["output_columns_names"]]
            elif isinstance(obj["output_columns_names"], list):
                self.output_columns_names = obj["output_columns_names"]
            else:
                raise ValueError(f'output_property or output_columns_names should be a list or a string')
        elif "output_property" in obj.keys() and obj["output_property"] is not None \
                and obj["output_property"] != "result":
            if isinstance(obj["output_property"], str):
                self.output_columns_names = [obj["output_property"]]
            elif isinstance(obj["output_property"], list):
                self.output_columns_names = obj["output_property"]
            else:
                raise ValueError(f'output_property or output_columns_names should be a list or a string')

        # source
        if "source" in obj.keys() and obj["source"] is not None:
            self.source = obj["source"]
            if self.source not in ["wizata", "mlflow"]:
                raise ValueError("source must be wizata or mlflow")

        # key and target
        if "model_key" not in obj.keys() or obj["model_key"] is None:
            raise KeyError('model_key must be declared in the config.')
        self.model_key = obj["model_key"]
        if "by_twin" in obj.keys():
            self.by_twin = get_bool(obj, name="by_twin")
        if "by_property" in obj.keys():
            self.by_property = get_bool(obj, name="by_property")
        if "property_name" in obj.keys() and obj["property_name"] is not None:
            self.property_name = obj["property_name"]
        if "model_type" in obj.keys() and obj["model_type"] is not None:
            self.model_type = obj["model_type"]
        if "model_alias" in obj.keys() and obj["model_alias"] is not None:
            self.model_alias = obj["model_alias"]

    def to_json(self, target: str = None):
        obj = {
            "source": self.source
        }

        # training info
        if self.train_script is not None:
            obj["train_script"] = str(self.train_script)
        if self.train_test_split_pct is not None:
            obj["train_test_split_pct"] = float(self.train_test_split_pct)
        if self.train_test_split_type is not None:
            obj["train_test_split_type"] = str(self.train_test_split_type)
        if self.features is not None:
            obj["features"] = self.features
        if self.properties_mapping is not None and isinstance(self.properties_mapping, dict):
            obj["properties_mapping"] = self.properties_mapping

        # features info
        if self.target_feat is not None:
            obj["target_feat"] = self.target_feat
        if self.output_append is not None:
            obj["output_append"] = str(self.output_append)
        if self.output_columns_names is not None:
            obj["output_columns_names"] = self.output_columns_names
        if self.output_prefix is not None:
            obj["output_prefix"] = self.output_prefix
        if self.function is not None:
            obj["function"] = self.function

        # key and target
        if self.model_key is not None:
            obj["model_key"] = self.model_key
        if self.model_type is not None:
            obj["model_type"] = self.model_type
        if self.model_alias is not None:
            obj["model_alias"] = self.model_alias
        if self.by_twin is not None:
            obj["by_twin"] = str(self.by_twin)
        if self.by_property is not None:
            obj["by_property"] = str(self.by_property)
        if self.property_name is not None:
            obj["property_name"] = self.property_name

        return obj

    def has_target_feat(self) -> bool:
        """
        determine if configuration possess a target feature
        """
        if self.target_feat is None:
            return False

        if isinstance(self.target_feat, str):
            return True
        elif isinstance(self.target_feat, list):
            if len(self.target_feat) == 0:
                return False
            else:
                return True
        else:
            raise TypeError(f'unsupported target_feat type {self.target_feat.__class__.__name__}')


class MLModel(ApiDto):
    """
    used to define a stored machine learning model within Wizata.
    - deprecated -
    """

    @classmethod
    def route(cls):
        return "mlmodels"

    @classmethod
    def from_dict(cls, data):
        obj = MLModel()
        obj.from_json(data)
        return obj

    @classmethod
    def get_type(cls):
        return "pickle"

    def __init__(self,
                 model_id: str = None,
                 generated_by_id=None,
                 exact_names=True,
                 exact_numbers=True,
                 key=None):
        self.model_id = model_id
        self.key = key

        self.generatedById = generated_by_id

        self.status = 'draft'
        self.input_columns = []
        self.output_columns = []

        self.needExactColumnNumbers = exact_numbers
        self.needExactColumnNames = exact_names
        self.has_anomalies = False
        self.label_counts = 0
        self.has_target_feat = False

        self.trained_model = None
        self.scaler = None

        self.experimentId = None

    def api_id(self) -> str:
        return str(self.model_id).upper()

    def endpoint(self) -> str:
        return "MLModels"

    def to_json(self, target: str = None):
        obj = {"id": str(self.model_id),
               "status": str(self.status),
               "needExactColumnNames": str(self.needExactColumnNames),
               "needExactColumnNumbers": str(self.needExactColumnNumbers),
               "hasAnomalies": str(self.has_anomalies),
               "hasTargetFeat": str(self.has_target_feat),
               "labelCount": str(self.label_counts)
               }
        if self.key is not None:
            obj["key"] = str(self.key)
        if self.generatedById is not None:
            obj["generatedById"] = self.generatedById
        if self.input_columns is not None:
            obj["inputColumns"] = json.dumps(list(self.input_columns))
        if self.output_columns is not None:
            obj["outputColumns"] = json.dumps(list(self.output_columns))
        if self.experimentId is not None:
            obj["experimentId"] = str(self.experimentId)
        return obj

    def from_json(self, obj):
        if "id" in obj.keys():
            self.model_id = obj["id"]
        if "key" in obj.keys() and obj["key"] is not None:
            self.key = obj["key"]
        if "generatedById" in obj.keys() and obj["generatedById"] is not None:
            self.generatedById = int(obj["generatedById"])
        if "experimentId" in obj.keys() and obj["experimentId"] is not None:
            self.experimentId = uuid.UUID(obj["experimentId"])
        if "status" in obj.keys():
            self.status = str(obj["status"]).lower()
        if "inputColumns" in obj.keys():
            self.input_columns = json.loads(obj["inputColumns"])
        if "outputColumns" in obj.keys():
            self.output_columns = json.loads(obj["outputColumns"])
        if "labelCount" in obj.keys():
            self.label_counts = int(obj["labelCount"])
        if "hasAnomalies" in obj.keys():
            if isinstance(obj["hasAnomalies"], str) and obj["hasAnomalies"].lower() == "false":
                self.has_anomalies = False
            else:
                self.has_anomalies = bool(obj["hasAnomalies"])
        if "hasTargetFeat" in obj.keys():
            if isinstance(obj["hasTargetFeat"], str) and obj["hasTargetFeat"].lower() == "false":
                self.has_target_feat = False
            else:
                self.has_target_feat = bool(obj["hasTargetFeat"])
        if "needExactColumnNumbers" in obj.keys():
            if isinstance(obj["needExactColumnNumbers"], str) and obj["needExactColumnNumbers"].lower() == "false":
                self.needExactColumnNumbers = False
            else:
                self.needExactColumnNumbers = bool(obj["needExactColumnNumbers"])
        if "needExactColumnNames" in obj.keys():
            if isinstance(obj["needExactColumnNames"], str) and obj["needExactColumnNames"].lower() == "false":
                self.needExactColumnNames = False
            else:
                self.needExactColumnNames = bool(obj["needExactColumnNames"])

    def get_sample_payload(self):
        pl_columns = {"timestamp": "[timestamp]"}
        for hardwareId in self.input_columns:
            pl_columns[hardwareId] = "[" + hardwareId + "]"
        pl_json = {
            "id": str(self.model_id),
            "dataset": pl_columns
        }
        return pl_json
